from collections import defaultdict

from .corpus import Corpus


class Analysis:
    def __init__(self, obj):
        self.corpora = {section: Corpus(section, text) for section, text in obj.corpora().items()}
        self.results = {}
        self.summary = defaultdict(int)

    def analyse(self, keywords, **kwargs):
        for kw in keywords:
            result = self.check_for_keyword(kw, **kwargs)
            self.results[kw] = result
            for section in result:
                self.summary[kw] += len(result[section])

    def check_for_keyword(self, keyword, **kwargs):
        return {
            section: c.delemmatized_concordance_list(keyword, **kwargs) 
            for section, c in self.corpora.items()
        }


from csv import DictWriter

from .analysis import Analysis

class CSVExporter:
    def __init__(self, file, output_path):
        self.file = file
        self.output_path = output_path / f"{self.file.type}s"
        self.output_path.mkdir(exist_ok=True)
        self.code_header = f"{self.file.type} code"
        self.name_header = f"{self.file.type} full title"
        self.summary_path = self.output_path / 'summary.csv'

    def export(self, keywords):
        with self.summary_path.open('w') as summary_file:
            summary_csv = DictWriter(summary_file, fieldnames=[self.code_header, self.name_header, *keywords])
            summary_csv.writeheader()
            for obj in self.file:
                analysis = Analysis(obj)
                analysis.analyse(keywords)
                summary_record = analysis.summary
                total_matches = sum(summary_record.values())
                summary_record[self.code_header] = obj.code
                summary_record[self.name_header] = obj.full_title
                summary_csv.writerow(summary_record)
                if not total_matches:
                    continue
                detail_path = self.output_path / f"{obj.code}.txt"
                with detail_path.open('w') as detail_file:
                    for kw, results in analysis.results.items():
                        if not summary_record[kw]:
                            continue
                        for section, examples in results.items():
                            if not examples:
                                continue
                            msg = f'found keyword "{kw}" {len(examples)} times in "{section}"'
                            print(f"{'='*len(msg)}\n{msg}\n{'='*len(msg)}\n", file=detail_file)
                            print("\n\n".join([f"'{c}'" for c in examples]), "\n", file=detail_file)



"""Corpus instances are initialised with text and can calculate everything we need"""

import nltk
from nltk.stem import WordNetLemmatizer

nltk.download('wordnet', quiet=True)
lemmatizer = WordNetLemmatizer()

def slice_from_concordance(con):
    return slice(
        con.offset - len(con.left),
        con.offset + len(con.right) + 1
    )

def delemmatize(concordance, tokens):
    t = tokens[slice_from_concordance(concordance)]
    i = len(concordance.left)
    t[i] = t[i].upper()
    return " ".join(t)

class Corpus:
    def __init__(self, label, text):
        self.label = label
        self.raw_string = text
        self.tokens = [word.lower() for sent in nltk.sent_tokenize(text) for word in nltk.word_tokenize(sent)]
        self.lemmatized_tokens = [lemmatizer.lemmatize(t) for t in self.tokens]
        self.text = nltk.text.Text(self.tokens)
        self.lemmatized_text = nltk.text.Text(self.lemmatized_tokens)
        self.vocab = self.text.vocab()
        self.lemmatized_vocab = self.lemmatized_text.vocab()

    def delemmatized_concordance_list(self, kw, **kwargs):
        lemmatized = self.lemmatized_text.concordance_list(kw, **kwargs)
        return [delemmatize(l, self.tokens) for l in lemmatized]

    def __repr__(self):
        return f"Corpus({self.label})"




"""Details of the structure of txt files, extract data into objects with properties"""

def extractData(path, split_string):
    with path.open('r', encoding='utf8') as f:
        data = f.read()
        chunks = data.split(split_string)
        chunks = [ch.splitlines() for ch in chunks]
        # chunks = [[line.strip() for line in ch if line.strip()] for ch in chunks]
        chunks[0] = ['deleteme', '', *chunks[0]]
        chunks = [ch[1:] for ch in chunks][:-1]
        for chunk in chunks:
            yield chunk


def file_factory(path):
    with path.open('r', encoding='utf8') as f:
        header = f.readlines()[2].strip()
    assert header in ["Module Specification", "Programme Specification"]
    if header == "Module Specification":
        return ModuleFile(path)
    else:
        return ProgrammeFile(path)



def aggregate_until(data, end_item):
    result = []
    while True:
        if data[0].startswith(end_item):
            break
        row = data.pop(0).strip()
        if not row:
            continue
        result.append(row)
    return result


class Programme:
    def __init__(self, data):
        assert data[3] == "Programme Specification"
        assert data[9] == "Programme Full Title: "
        assert data[15] == "Programme Short Title: "
        self.full_title = data[11]
        self.short_title = data[13]
        assert data[17] == "Programme Code: "
        assert data[19] == "Programme Type: "
        self.type = data[21]
        self.code = data[23]

        data = data[24:]
        self.something = "\n".join(aggregate_until(data, "Faculty: "))
        assert data[0] == "Faculty: "
        self.faculty = data[2]
        assert data[4] == "School: "
        self.faculty = data[6]
        assert data[8] == "Department: "
        self.department = data[10]
        assert data[12] == "Programme Leader: "
        self.programme_leader = data[14]
        assert data[16] == "Mode of Delivery: "
        assert data[18] == "Normal Duration: "
        self.mode = data[22]
        self.duration = data[20]
        assert data[24] == "Offered at the following sites:"

        data = data[25:]
        self.sites = aggregate_until(data, "Distance Learning availability:")
        assert data[0] == "Distance Learning availability: "
        _ = aggregate_until(data, 'Awards Available:')

        assert data[0] == "Awards Available:\t"
        award_keys = data[1].split('\t')
        data = data[2:]
        awards = aggregate_until(data, 'Relevant QAA subject benchmarking statement(s):')
        self.awards = [{k: v for k, v in zip(award_keys, a.split('\t'))} for a in awards]

        assert data[0] == "Relevant QAA subject benchmarking statement(s):"
        data = data[1:]
        self.qaa_benchmarking_statements = aggregate_until(data, 'Accreditation Details:')
        data = data[1:]
        self.accreditation_details = aggregate_until(data, "Entry Requirements")
        self.entry_requirements = aggregate_until(data, 'Programme Description')

        assert data[0] == "Programme Description: Characteristics and Aims"
        self.description = data[1].strip()
        assert data[3] == "Learning, Teaching and Assessment Strategies:"
        self.learning_teaching_and_assessment_strategies = data[4].strip()
        assert data[6] == "Structure and Regulations:"
        self.structure_and_regulations = data[7].strip()
        assert data[9] == "Learning Outcomes:"
        self.learning_outcomes = data[10].strip()
        data = data[12:]
        assert data[0].strip() == 'Module Details (across all module groups):'

        keys = data[1].split('\t')
        data = data[2:]
        values = aggregate_until(data, 'Any programme-specific differences or regulations:')
        self.modules = [{k: v for k, v in zip(keys, v.split('\t'))} for v in values]
        assert data == ['Any programme-specific differences or regulations:', '', '']

    def __str__(self):
        return f"Programme({self.code})"

    def corpora(self):
        return {
            "description": self.description,
            "learning_teaching_and_assessment_strategies": self.learning_teaching_and_assessment_strategies,
            "learning outcomes": self.learning_outcomes
        }


class Module:
    def __init__(self, data):
        assert data[3] == "Module Specification"
        assert data[9] == "Module Full Title: "
        assert data[15] == "Module Short Title: "
        assert data[17] == "Module Code: "
        assert data[19] == "Credit Value: "
        assert data[23] == "Credit Level: "
        assert data[29] == "Faculty: "
        assert data[33] == "School: "
        assert data[37] == "Department: "
        assert data[41] == "Module Leader: "
        assert data[45] == "Module Appraiser(s) / Marker(s): "

        self.full_title = data[11]
        self.short_name = data[13]
        self.credit_value = data[21]
        self.credit_level = data[25]
        self.code = data[27]
        self.faculty = data[31]
        self.school = data[35]
        self.department = data[39]
        self.module_leader = data[43]
        data = data[46:]

        self.module_appraisers = aggregate_until(data, "Offered at the following sites:")
        assert data[0] == "Offered at the following sites:"
        data = data[1:]

        self.sites = aggregate_until(data, "Semester / Year-long:")
        assert data[0] == "Semester / Year-long: "
        data = data[1:]

        self.intakes = aggregate_until(data, "DMUGlobal Content indicator:")

        assert data[0] == "DMUGlobal Content indicator: "
        assert data[2] in ["Y", "N"]
        self.DMUGlobal = data[2] == "Y"

        assert data[4] == "Ethical approval required: "
        assert data[6] in ["Y", "N"]
        self.ethical_approval = data[6] == "Y"

        assert data[8] == "Details of Accreditation by Professional, Statutory or Regulatory Body:"
        data = data[9:]
        self.accreditation = aggregate_until(data, "Module Pre-requisites:")
        assert data[0] == "Module Pre-requisites:"
        data = data[1:]

        self.prerequisites = aggregate_until(data, "Module Description:")
        assert data[0] == "Module Description:"
        assert data[3] == "Learning Outcomes:"

        self.description = data[1]
        self.learning_outcomes = data[4]

        assert data[6].strip() == "Evaluation:"

        assessment_keys = data[7].split('\t')
        data = data[8:]
        assert len(assessment_keys) == 8
        assessments = aggregate_until(data, "Anonymous marking exemption codes:")
        self.assessments = [{k: v for k, v in zip(assessment_keys, a.split('\t'))} for a in assessments]
        assert data[
                   0] == "Anonymous marking exemption codes: OPTO1: Individually distinct work; OPTO2: Reflection on development of own work; OPTO3:"

        assert data[3] == "Assessment Notes:"
        self.assessment_notes = data[4]

        assert data[6] == "Reassessment:"
        self.reassessment = data[7]

        assert data[9] == "Expected Methods of Delivery:"
        self.method_of_delivery = data[10]

        assert data[12].strip() == "Programmes using this module:"
        programme_keys = data[15].split('\t')
        data = data[16:]
        programmes = aggregate_until(data, "Remarks:")
        self.programmes = [{k: v for k, v in zip(programme_keys, p.split('\t'))} for p in programmes]
        # assert data == ["Remarks:", "", ""]
        assert data[0].strip() == "Remarks:"
        self.remarks = [remark for remark in data[1:] if remark]

    def corpora(self):
        """
        This method should return the main corpus for this module.
        Make adjustments here to control what we actually analyse.
        """
        return {
            "description": self.description
        }

    def __str__(self):
        return f"Module({self.code})"


class DataFile:
    stop = None    

    def __init__(self, path):
        self.path = path

    def __iter__(self):
        for d in extractData(self.path, self.stop):
            yield self.item_class(d)


class ProgrammeFile(DataFile):
    stop = 'End of Programme Specification for'
    type = 'programme'
    item_class = Programme


class ModuleFile(DataFile):
    stop = 'End of Module Specification for'
    type = 'module'
    item_class = Module



import json
from pathlib import Path
from distutils.dir_util import copy_tree

from .analysis import Analysis

class JSONExporter:
    def __init__(self, file, output_path):
        self.file = file
        self.output_path = output_path
        self.output_path.mkdir(exist_ok=True)
        self.code_header = f"{self.file.type} code"
        self.name_header = f"{self.file.type} full title"
        self.summary_path = self.output_path / 'summary.json'

    def export(self, keywords):
        result = []
        for obj in self.file:
            analysis = Analysis(obj)
            analysis.analyse(keywords)
            record = {
                "code": obj.code,
                "title": obj.full_title,
                "data": analysis.results
            }
            result.append(record)
        with self.summary_path.open('w') as summary_file:
            json.dump(result, summary_file)
        copy_tree(str(Path(__file__).parent / 'html'), str(self.output_path), update=True)


"""This is the main command line interface"""

from configparser import ConfigParser
from pathlib import Path
from argparse import ArgumentParser

from .keywords import load_keywords_file
from .txt_parser import file_factory
from .csv_exporter import CSVExporter
from .json_exporter import JSONExporter

exporters = {
    'csv': CSVExporter,
    'json': JSONExporter,    
}

def main(filename, conf):
    # check the provided config file exists
    conf_path = Path(conf).expanduser()
    if not conf_path.exists():
        print(f"Wait! We are missing the configuration file at {conf_path}")
        default = (Path(__file__).parent / 'config.cfg.default').read_text()
        conf_path.write_text(default);
        print(f"We added this configuration to the file for you.\n")
        print(default)
        print(f"For now, we are aborting.")
        print(f"Please check/edit the configuration before running again.")
        exit()

    # load the config file
    config = ConfigParser()
    config.read(conf_path)

    # Load keywords
    keyword_path = Path(config.get("curriculumAnalysis", "keywords_path")).expanduser()
    if not keyword_path.exists():
        default = (Path(__file__).parent / 'keywords.txt').read_text()
        keyword_path.write_text(default);
        
    keywords = load_keywords_file(keyword_path)

    # load the data
    file = file_factory(Path(filename).expanduser())

    # Generate an analysis
    outpath = Path(config.get("curriculumAnalysis", "outpath")).expanduser()
    format = config.get("curriculumAnalysis", "format")
    exporter = exporters[format](file, outpath)

    # Export it
    exporter.export(keywords)

def cli():
    # The parser accepts a filename and an optional configuration file argument
    parser = ArgumentParser(epilog="For more information see https://github.com/IESD/curriculumAnalysis", description='A simple tool for analysing DMU module and programme specifications with respect to provided keywords.')
    parser.add_argument('filename', help='the file to process')
    parser.add_argument('-c', '--conf', default="~/.curriculumAnalysis/config.cfg", help='configuration file (default ~/.curriculumAnalysis/config.cfg)')
    args = parser.parse_args()
    main(args.filename, args.conf)


if __name__ == "__main__":
    cli()